from .ZipFilesTree import ZipFilesTree, ZipFilesManager
from .DirectoryFilesTree import DirectoryFilesTree
def files_tree(target, replace=False):
    """
    Parameters
    ----------

    target
      Either the path to a target folder, or a zip file, or '@memory' to write
      a zip file in memory (at which case a string of the zip file is returned)

    replace
      If True, will remove the target if it already exists. If False, new files
      will be written inside the target and some files may be overwritten.
    """
    if not isinstance(target, str):
        return ZipFilesTree(files_manager=ZipFilesManager(source=target))
    elif target == '@memory':
        return ZipFilesTree("@memory")
    elif target.lower().endswith(".zip"):
        return ZipFilesTree(target)
    else:
        try:
            return DirectoryFilesTree(target)
        except:
            pass

    # Last chance, lets try again:
    return ZipFilesTree(files_manager=ZipFilesManager(source=target))



import os
import re

expr = re.compile(r"[^a-zA-Z\d]")
def sanitize_name(name):
    if name[0] in "0123456789":
        name = "_" + name
    return re.sub(expr, "_", name)

class FilesTree:
    _is_dir = True

    def __init__(self, location=".", name=None, files_manager=None,
                 replace=False):

        # Initialize the properties and the files manager
        self._name = name
        self._location = location
        self._dict = {}
        if name is None:
            self._path = self._location
        else:
            self._path = os.path.join(self._location._path, self._name)
        if files_manager is None:
            files_manager = self._init_files_manager()
        self._files_manager = files_manager
        self._files_manager.create(self, replace=replace)

        # Automatically explore the folder and subfolders to build a tree
        self._files = []
        self._dirs = []
        if self._is_dir:
            for filename in self._files_manager.list_files(self):
                self._file(filename, replace=False)
            for dirname in self._files_manager.list_dirs(self):
                self._dir(dirname, replace=False)

    def _dir(self, name, replace=True):
        if name not in self._dict:
            subdir = FilesTree(location=self, name=name,
                               files_manager=self._files_manager,
                               replace=replace)
            self._dirs.append(subdir)
            self._dict[name] = subdir
            self.__dict__[sanitize_name(subdir._name)] = subdir
        return self._dict[name]

    def _file(self, name, replace=True):
        if name not in self._dict:
            f = File(location=self, name=name,
                     files_manager=self._files_manager,
                     replace=replace)
            self._files.append(f)
            self._dict[f._name] = f
            self.__dict__[sanitize_name(f._name)] = f
        return self._dict[name]

    def _delete(self):
        if isinstance(self._location, str):
            raise IOError("Looks like you are trying to delete the root.")
        self._files_manager.delete(self)
        location = self._location
        location._dict.pop(self._name)
        location.__dict__.pop(self._name, None)
        if self._is_dir:
            location._dirs = [subdir for subdir in location._dirs
                              if subdir._name == self._name]
        else:
            location._files = [f for f in location._dirs
                               if f._name == self._name]

    def __getitem__(self, it):
        return self._dict[it]

    @property
    def _all_files(self):
        return self._files + sum([d._all_files for d in self._dirs], [])

    def _tree_view(self, indent_size=2, indent_level=0, as_lines=False):
        """Return a string representation of the tree for pretty printing.
        """
        lines = []
        space = indent_size * indent_level * " "
        for subdir in sorted(self._dirs, key=lambda subdir: subdir._name):
            lines.append("%s%s/" % (space, subdir._name))
            lines += subdir._tree_view(indent_size=indent_size,
                                       indent_level=indent_level + 1,
                                       as_lines=True)
        for f in sorted(self._files, key=lambda f: f._name):
            lines.append("%s%s" % (space, f._name))

        return lines if as_lines else "\n".join(lines)

    def _close(self):
        pass

    def __exit__(self, *a):
        self._close()

    def __enter__(self):
        return self


class File(FilesTree):
    _is_dir = False

    def read(self, mode="r"):
        return self._files_manager.read(self, mode=mode)

    def write(self, content, mode="a"):
        if hasattr(content, "decode"):
            mode += "b"
        self._files_manager.write(self, content, mode=mode)



import os
import shutil
from .FilesTree import FilesTree


class DirectoryFilesManager:

    @staticmethod
    def list_files(directory):
        path = directory._path
        return [] if not os.path.exists(path) else [
            name for name in os.listdir(path)
            if os.path.isfile(os.path.join(path, name))
        ]
    @staticmethod
    def list_dirs(directory):
        path = directory._path
        return [] if not os.path.exists(path) else [
            name for name in os.listdir(path)
            if os.path.isdir(os.path.join(path, name))
        ]

    @staticmethod
    def read(fileobject, mode="r"):
        with open(fileobject._path, mode=mode) as f:
            result = f.read()
        return result

    @staticmethod
    def write(fileobject, content, mode="a"):
        with open(fileobject._path, mode=mode) as f:
            f.write(content)

    @staticmethod
    def delete(directory):
        if directory._is_dir:
            shutil.rmtree(directory._path)
        else:
            os.remove(directory._path)

    def create(self, directory, replace=True):
        path = directory._path
        if replace and os.path.exists(path):
            self.delete(directory)
        if replace or (not os.path.exists(path)):
            if directory._is_dir:
                os.mkdir(path)
            else:
                with open(path, "w") as f:
                    pass

    @staticmethod
    def join_paths(*paths):
        return os.path.join(*paths)

class DirectoryFilesTree(FilesTree):

    def _init_files_manager(self):
        return DirectoryFilesManager()



import os
import sys
import zipfile
from collections import defaultdict
from .FilesTree import FilesTree
import re

PYTHON3 = (sys.version_info[0] == 3)

if PYTHON3:
    from io import StringIO, BytesIO
    StringBytesIO = BytesIO
else:
    from StringIO import StringIO
    StringBytesIO = StringIO



class ZipFilesManager:

    def __init__(self, path=None, source=None):
        if source is None:
            if path == "@memory":
                source = StringBytesIO()
            else:
                source = path
        elif isinstance(source, (str, bytes)):
            source = StringBytesIO(source)
        self.source = source
        self.path = "." if path is None else path

        print ("asdsadasd", source)
        #print ("ddd", os.path.exists(source))
        if path == "@memory":
            # Horrible hack: we give an empty but valid zip to the reader:
            reader_source = StringBytesIO(b'PK\x05\x06' + 18 * b'\x00')
            self.reader = zipfile.ZipFile(reader_source, "r")

        elif isinstance(source, str) and not os.path.exists(source):
            print ("IM here")
            reader_source = StringBytesIO(b'PK\x05\x06' + 18 * b'\x00')
            self.reader = zipfile.ZipFile(reader_source, "r")
        else:
            print ("Im there")
            self.reader = zipfile.ZipFile(source, "r")
        self.files_data = defaultdict(lambda *a: StringBytesIO())
        self.writer = zipfile.ZipFile(source, "a",
                                      compression=zipfile.ZIP_DEFLATED)

    def relative_path(self, directory):
        path = directory._path[len(self.path)+1:]
        if directory._is_dir and path != "":
            path += "/"
        return path


    def list_files(self, directory):
        path = self.relative_path(directory)
        matches = [
            re.match(r"%s([^/]*)$" % path, name)
            for name in self.reader.namelist()
        ]
        files = set([
            match.groups()[0]
            for match in matches
            if match is not None
            and (match.groups()[0] != "")
        ])
        return files


    def list_dirs(self, directory):
        path = self.relative_path(directory)
        matches = [
            re.match(r"%s([^/]*)/" % path, name)
            for name in self.reader.namelist()
        ]
        dirs = set([
            match.groups()[0]
            for match in matches
            if match is not None
            and (match.groups()[0] != "")
        ])
        return dirs


    def read(self, fileobject, mode="r"):
        path = self.relative_path(fileobject).strip("/")
        if path in self.files_data:
            return self.files_data[path].getvalue()
        else:
            return self.reader.read(path)

    def write(self, fileobject, content, mode="w"):
        path = self.relative_path(fileobject)
        if self.path_exists_in_file(fileobject):
            raise NotImplementedError(
                "Rewriting a file already zipped is not currently supported. "
                "It may actually not even be possible, or in an inelegant way."
            )
        if mode in ("w", "wb"): # i.e. not append
            self.files_data.pop(path) # overwrite !
        if not isinstance(content, bytes):
            content = content.encode("utf-8")
        self.files_data[path].write(content)

    def delete(self, directory):
        raise NotImplementedError(
            "Deleting/modifying/overwriting an already-zipped file "
            "is not currently supported. "
            "It may actually not even be possible, or in a very inelegant way."
        )

    def create(self, directory, replace=False):
        if self.path_exists_in_file(directory) and replace:
            self.delete(directory)
        # TODO: I don't know how to create an empty dir in a zip but right
        # now it doesn't really matter because the directories are created
        # the moment we create a file whose address in this directory.


    def path_exists_in_file(self, directory):
        print (self.reader)
        print (self.reader.namelist())
        print (self.relative_path(directory))

        return self.relative_path(directory) in self.reader.namelist()


    @staticmethod
    def join_paths(*paths):
        return "/".join(*paths)

    def close(self):
        for path, data in self.files_data.items():
            self.writer.writestr(path, data.getvalue())
        self.writer.close()
        if hasattr(self.source, "getvalue"):
            return self.source.getvalue()


class ZipFilesTree(FilesTree):
    """
    >>> ZipFilesTree("some_archive.zip")
    >>> # Data is either a sting/bytes, or a file-like object.
    >>> ZipFilesTree(files_manager=ZipFilesManager(source=some_data))
    """

    def _init_files_manager(self):
        return ZipFilesManager(self._path)

    def _close(self):
        return self._files_manager.close()
